/*
 * main.c -- the core of pxdrv
 *
 * Copyright (C) 1997   rubini@linux.it (Alessandro Rubini)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********/

/*
 * Missing features:
 *      trigger control
 *      size change
 *      cache control
 *      camera selection
 *      debounce control
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__ /* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

char kernel_version [] = UTS_RELEASE;

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/bios32.h> /* pcibios_* */
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/io.h>

#include "px500.h"        /* local definitions */
#include "allocator.h"

/*
 * I don't use static symbols here, because modularization code hides
 * public symbols
 */

int px_major =   PX_MAJOR;

Px_Dev *px_devices; /* allocated in init_module */

/*
 * Different minors behave differently, so let's use multiple fops
 */
extern struct file_operations px_bin_fops;
extern struct file_operations px_pgm_fops;
extern struct file_operations px_ctl_fops;
extern struct file_operations px_hib_fops;
extern struct file_operations px_hip_fops;

struct file_operations *px_fop_array[]={
    &px_bin_fops,      /* type 0: binary */
    &px_pgm_fops,      /* type 1: pgm */
    &px_ctl_fops,      /* type 2: control */
    &px_hib_fops,      /* type 3: hires binary */
    &px_hip_fops       /* type 4: hires pgm */
};

#define PX_MAX_TYPE 4

    
u32 px_running_cfg[]= {
          0x000A0000, /* even field DMA address (?) */
          0x00000000,
          0x00000000,
          0x000A0A00, /* odd field DMA (?) */

/* 10 */  0x00000000,
          0x00000000,
          0x00000F00, /* even field stride */
          0x00000000,

/* 20 */  0x00000000,
          0x00000F00, /* odd field stride */
          0x00000000,
          0x00000000,

/* 30 */  0xEEEEEE01, /* 16 bit even fields */
          0xEEEEEE01, /* 16 bit odd fileds */
          0x00200020, /* FIFO trigger # Packed,Planar */
          0x00000003,

/* 40 */  0x000080F0, /* Capture control. 80F3 would means capture is on. */
          0x00000000,
          0x00000000, /* interrupts (disabled) */
          0x00000001, /* Field Mask, even.  Capture first frame in sequence. */

/* 50 */  0x00000001, /* Field mask, odd. */
          0x00000000, /* Mask length even,odd.  Sequence is (0+1)=1 frames */
          0x0005007C, /* FIFO almost full, empty. */
          0x01010101, /* I2C clock durations */

/* 60 */  0x00000300, /* I2C mode */
          0x00000000, /* I2C manual control */
          0x00000000, /* registers for I2C auto mode. */
          0x00000000,

/* 70 */  0x00000000,
          0x00000000,
          0x00000000,
          0x00000000,

/* 80 */  0x00000000,
          0x00000000,
          0x00000000,
          0x00000000, /* DMA end address even */

/* 90 */  0x00000000  /* DMA end odd field */

#define PX_CFG_ITEMS 37 /* 0x94/sizeof(u32) */
};

/*---------------------------------------------------------- wait for port */
void px_wait_port(Px_Dev *d)
{
    unsigned long j = jiffies + HZ/10;
    while (PX_READ32(d, PX_REG_SERDATA) & (1<<24))
        if (jiffies > j) { /* max 100ms */
            printk(KERN_WARNING "pxdrv: lockup accessing ports\n");
	    PDEBUG("Lockup from %p\n",__builtin_return_address(0));
            break; /* Hmm.... */
        }
}

/*---------------------------------------------------------- command */
static int px_partial_command(Px_Dev *d, int byte)
{
    int res;
    unsigned long j = jiffies + HZ/5;

    res = px_set_port_wait(d, PX_PORT_MICRO, byte);
    if (res) return res; /* error */
    do {
        res = px_get_port(d, PX_PORT_STATUS);
        if (res < 0) return res; /* error */
        if (jiffies > j) {
            printk(KERN_WARNING "pxdrv: lockup in command %02x\n", byte);
            return -EIO;
        }
    } while (!(res & PX_BIT_MP_ACK));
    res = px_get_port(d, PX_PORT_MICRO);
    PDEBUG("partial write cmd 0x%02x: read %02x\n",byte,res);
    return res;
}

int px_wr_command(Px_Dev *d, int command, unsigned long arg)
{
    int result, nextbyte=command;
    int count = 3; /* max nr failures */
    while (count--) {
        result = px_partial_command(d, nextbyte);
	switch(result) {
	    case 0:
	        return 0; /* done */
            case 1: case 2: case 3: case 4:
	        /* send the another byte */
		nextbyte = 0xff & (arg >> (8 * (result-1))); /* lsb first */
		count++; /* not failed */
		continue;
            default:
	        printk(KERN_WARNING "pxdrv: command returned 0x%02x\n",
		       result);
                if (result < 0) return result; /* error */
		/* else continue */
	}
    }
    return -EIO;
}

/* by now, only one-byte commands */
unsigned long px_rd_command(Px_Dev *d, int command)
{
    return px_partial_command(d,command);
}

/*---------------------------------------------------------- gain */
int px_set_gain(Px_Dev *dev, Px_Gain *ptr)
{
    int res = 0;
    if (ptr->range > 3) return -EINVAL;

    /* the delay between set_port() are needed */
    res += px_set_port_wait(dev, PX_PORT_ADDRESS, 0);
    udelay(100);
    res += px_set_port_wait(dev, PX_PORT_LEVEL,
                            0x3C | (0xC0 & ~(ptr->range << 6)));
    udelay(100);
    res += px_set_port_wait(dev, PX_PORT_ADDRESS, 2);
    udelay(100);
    res += px_set_port_wait(dev, PX_PORT_LEVEL, ptr->fine);

    if (res)
        return -EIO;
    dev->gain = *ptr;
    return 0;
}

inline int px_get_gain(Px_Dev *dev, Px_Gain *ptr)
{
    PDEBUG("get gain: %i %i\n", dev->gain.range, dev->gain.fine);
    *ptr = dev->gain;
    return 0;
}

/* using the gain as an int value is handy in autogain */

inline int px_get_intgain(Px_Dev *dev)
{
    return (dev->gain.range << 8) | dev->gain.fine;
}

inline int px_set_intgain(Px_Dev *dev, int gain)
{
    Px_Gain splitgain = {gain>>8, gain&0xff};
    return px_set_gain(dev, &splitgain);
}

/*---------------------------------------------------------- lut */
int px_set_lut(Px_Dev *dev)
{
    int i;
    px_set_port_wait(dev, PX_PORT_ADDRESS, 0);
    for (i=0; i<256; i++) {
        udelay(100);
        px_set_port_wait(dev, PX_PORT_DATA, dev->lut[i]);
    }
    return 0;
}

int px_reset_lut(Px_Dev *dev)
{
    int i;
    for (i=0; i<256; i++) {
        dev->lut[i] = i;
    }
    return px_set_lut(dev);
}

/*---------------------------------------------------------- up/down */
/*
 * When a grabber is first used, it must be powered on. Dma is still
 * kept disabled, but memory mapping is enabled
 */
int px_bringup(Px_Dev *d)
{
    u16 cmd; int i;

    if (d->isup) return 0;

    /* enable memory */
    pcibios_read_config_word(d->bus, d->device, PCI_COMMAND, &cmd);
    pcibios_write_config_word(d->bus, d->device, PCI_COMMAND,
                               cmd | PCI_COMMAND_MEMORY);

    /* write default cfg */
    for (i=0; i<PX_CFG_ITEMS; i++)
       PX_WRITE32(d, px_running_cfg[i], i*4);
    px_set_mode(d, PX_REG_GRAB_OFF);
    px_reset_ports(d);
    d->gain.range = 1; d->gain.fine = 0; px_set_gain(d, &d->gain);
    px_reset_lut(d);
    d->isup = 1;
    return 0;
}

int px_shutdown(Px_Dev *d)
{
    u16 cmd;

    d->isup = 0;
    pcibios_read_config_word(d->bus, d->device, PCI_COMMAND, &cmd);
    if ( (cmd & PCI_COMMAND_MEMORY) == 0) /* no memory: already off */
        return 0;
    px_set_mode(d, PX_REG_GRAB_NONE);
    PX_WRITE32(d, 0, PX_REG_DMAEND0);
    PX_WRITE32(d, 0, PX_REG_DMAEND1);
    if(d->irq_active) {
        free_irq(d->irq, d);
        d->irq_active = 0; /* safe */
    }

    /* disable memory */
    pcibios_write_config_word(d->bus, d->device, PCI_COMMAND,
                              cmd & ~PCI_COMMAND_MEMORY);
    return 0;
}

/*-------------------------------------------------------- init the device */
/*
 * Device initialization. Before calling this function, at power on,
 * the registers are all set to 0. The received pointer holds
 * valid bus, device and phys_addr fields
 */
int px_init(Px_Dev *d)
{
    u16 cmd; int i;

    d->regs = ioremap(d->phys_add, 4096); /* one page (on x86, at least) */
    if (!d->regs) return -ENOMEM;

    pcibios_read_config_byte(d->bus, d->device, PCI_INTERRUPT_LINE, &d->irq);
    PDEBUG("irq is %i\n",d->irq);

    /*
     * this overwrites registers, before checking the ID. But some of
     * the registers need to be set in order to access the ID. Bad.
     * At least, save its cfg status before acting on it.
     */
    pcibios_read_config_word(d->bus, d->device, PCI_COMMAND, &cmd);
    px_bringup(d);

    if (px_get_port(d, PX_PORT_ID) != 0xa5) { /* not correct id */
        printk(KERN_WARNING "px500: PCI device (%i,%i,%i) revealed not to "
	       "be a grabber after configuration\n",
	       d->bus, d->device>>3, d->device&7);
        px_shutdown(d);
	pcibios_write_config_word(d->bus, d->device, PCI_COMMAND, cmd);
        vfree((void *)(d->regs));
	d->regs=NULL;
	return -ENODEV;
    }

    /* well, write something useful... */
    for (i=0; i<PX_CFG_ITEMS; i++)
        PX_READ32(d, i*4) = px_running_cfg[i];
    px_set_mode(d, PX_REG_GRAB_NONE);

    /* restore cmd */
    pcibios_write_config_word(d->bus, d->device, PCI_COMMAND, cmd);
    px_shutdown(d);
    return 0; /* success */
}

/*---------------------------------------------------- the proc filesystem */

/*
 * The proc filesystem: function to read an entry
 */

int px_read_proc(char *buf, char **start, off_t offset,
                   int len, int unused)
{
    int i;
    Px_Dev *d;

    #define LIMIT (PAGE_SIZE-80) /* don't print anymore after this size */
    len=0;
    for(d=px_devices; d; d=d->next) {
        u32 dword;
        u16 cmd;
	u32 mode;

	if (d->isup == 0) px_bringup(d);
        d->usage++;

	if ( (mode = d->mode) == PX_REG_GRAB_NONE)
	    px_set_mode(d, PX_REG_GRAB_OFF);
        len = sprintf(buf,"\n\nDevice at %p (bus %i dev %i)\nPCI Registers:\n",
                    d, d->bus, d->device);
        for (i=0; i<0x40; i+=4) {
            if (i%16 == 0)
                len += sprintf(buf+len, "0x%02x: ",i);
            pcibios_read_config_dword(d->bus, d->device, i, &dword);
            len += sprintf(buf+len, "0x%08x%c", dword, (i+4)%16 ? ' ' : '\n');
            if (len > LIMIT) return len;
        }
        /* Save the current settings, enable memory mapping and snooze it */
        pcibios_read_config_word(d->bus, d->device, PCI_COMMAND, &cmd);
        pcibios_write_config_word(d->bus, d->device, PCI_COMMAND,
                               cmd | PCI_COMMAND_MEMORY);
        len += sprintf(buf+len, "\n7116 registers:\n");
        for (i=0; i<0xA0; i+=4) {
            if (i%16 == 0)
                len += sprintf(buf+len, "0x%02x: ", i);
            dword = ((volatile u32 *)(d->regs))[i/4];
            len += sprintf(buf+len, "0x%08x%c", dword, (i+4)%16 ? ' ' : '\n');
            if (len > LIMIT) break;
        }

        if (len < LIMIT) {
            len += sprintf(buf+len, "\nPorts:\n");
            for (i=PX_PORT_MIN; i <= PX_PORT_MAX; i++) {
                len += sprintf(buf+len, "0x%02x%c", px_get_port(d, i)&0xff,
                               i==PX_PORT_MAX ? '\n' : ' ');
                if (len > LIMIT) break;
            }
        }
        pcibios_write_config_word(d->bus, d->device, PCI_COMMAND, cmd);
	px_set_mode(d, mode);
        d->usage--;
        if (d->usage == 0) px_shutdown(d);
    }
    #undef LIMIT
    return len;
}

struct proc_dir_entry px_proc_entry = {
        0,                 /* low_ino: the inode -- dynamic */
        5, "pxdrv",        /* len of name and name */
        S_IFREG | S_IRUGO, /* mode */
        1, 0, 0,           /* nlinks, owner, group */
        0, NULL,           /* size - unused; operations -- use default */
        &px_read_proc,     /* function used to read data */
        /* nothing more */
    };

/*---------------------------------------------------------- allocate DMA */
void px_free_dma(Px_Dev *dev)
{
    if (dev->vdmabuffer) {
	iounmap(dev->vdmabuffer);
	dev->vdmabuffer=0;
    }
    if (dev->dmabuffer) {
        PDEBUG("freeing DMA buffer\n");
        allocator_free_dma(dev->dmabuffer);
        dev->dmabuffer =0;
        PDEBUG("freed DMA buffer\n");
    }

    dev->mapbuffer = 0;
    dev->usage = 0;
    dev->dmasize = dev->imagesize = 0;
    /* don't override other values, which are useful in reallocating */
}

int px_allocate_dma(Px_Dev *dev, int kilobytes)
{
    if (dev->dmasize >= kilobytes * 1024)
        return 0;
    if (dev->dmasize)
        px_free_dma(dev);
    dev->dmabuffer = allocator_allocate_dma(kilobytes,GFP_KERNEL);
    PDEBUG("allocating: %i kilos, got 0x%08lx\n", kilobytes,dev->dmabuffer);

    if (!dev->dmabuffer) {
#if 0
        allocator_free_map(dev->mapbuffer);
#endif
	dev->mapbuffer = 0;
	return -ENOMEM;
    }
    /* this is the one accessed by driver code */
    dev->dmasize = kilobytes * 1024;
    dev->vdmabuffer = ioremap(dev->dmabuffer, dev->dmasize);
    dev->usage = 0;
    return 0;
}

/*---------------------------------------------------- interrupt handler */
static void px_interrupt(int irq, void *devptr, struct pt_regs *regs)
{
    Px_Dev *dev=devptr;
    int status, compare;
    /*
     * First of all, we should check if the interrupt is valid.
     * A shared handler is invoked on any interrupt of any device,
     * so we must validate it.
     */
    
    status = px_get_port(dev, PX_PORT_STATUS);
    if ( (status & PX_BIT_IRQ_STATUS) == 0)
        return;

    px_set_control_bit(dev, PX_BIT_IRQ_ACK); /* ack + enable */
    dev->irq_count++;
    
    if (dev->isup < 3) {
      dev->isup++;
      PDEBUG("isup %i\n",dev->isup);
  }
    if (dev->queue && dev->isup == 3) {
        wake_up_interruptible(&dev->queue);
        PDEBUG("waking up\n");
    }

    if ( (dev->flags & PX_FLAG_AUTOGAIN) == 0)
        return;

    /* try to do autogain */

    compare = 1; /* lower, need raise by default */
    if (status & PX_BIT_GREATER) compare = -1;
    else if (status & PX_BIT_EQUAL) compare = 0;

    if ( !compare || (compare * dev->autogain_now) < 0) { /* ok or overrrun */
        dev->autogain_now = 0;
        return;
    }
    /* else, need do change, maybe strenghten the change */
    if ( (dev->autogain_now != dev->autogain_step) &&
	(dev->autogain_now != -(dev->autogain_step)) )
        dev->autogain_now += compare;
    compare = px_get_intgain(dev) + dev->autogain_now;
    if (compare < 0) compare = 0;
    if (compare > 1024) compare = 1023;
    px_set_intgain(dev, compare);
}


/*-------------------------------------------------------- setup dma */
/* Low level parameters */
int px_set_dma_parameters(Px_Dev *dev)
{
    int result = 0;
    int decimate = 0;
    unsigned long addr = virt_to_bus((void*)dev->dmabuffer);

    int i;
    /* Only set parameters on first open, to avoid conflicts */
    if (dev->usage) {
        PDEBUG("warning: re-set parameters: ignoring\n");
        return 0;
    }

    PDEBUG("parameters: %i, %i, %li\n",dev->firstline,
           dev->lastline, dev->dmahole);
    if (!dev->dmabuffer || !dev->xsize || (dev->dmahole&3))
        return -EINVAL;

    /*
     * Decimation: only 640, 320 etc are valid. Refuse other values by
     * now. (Although one might play it dirty and overwrite the end of
     * the previous line with the next one :)
     */
    i = dev->xsize;
    while (i < 640) {
        decimate++;
        i *= 2;
    }
    if (i != 640) {
        printk("px500: wrong x size\n");
        return -EINVAL;
    }
    PDEBUG("decimate %i\n", decimate);
    px_decimate(dev, decimate);

#if 0
    /* first, write something silly to it */
    {
    char *p = (char *) dev->vdmabuffer; /* Vdma */
    int i, j;
    for (i=0; i<dev->ysize; i++)
      for (j=0; j<dev->xsize; j++)
	*(p++)=i;
    }
#endif

    PX_WRITE32(dev, addr, PX_REG_DMA0);
    PX_WRITE32(dev, addr + dev->dmahole, PX_REG_DMA1);
    PX_WRITE32(dev, dev->dmahole,    PX_REG_HOLE0);
    PX_WRITE32(dev, dev->dmahole,    PX_REG_HOLE1);

    udelay(100); /* without this one, the command locks */
    
    result += px_wr_command(dev, PX_CMD_STARTLINE, dev->firstline);
    result += px_wr_command(dev, PX_CMD_ENDLINE, dev->lastline >> 2);
    if (result) { /* one or both failed */
        PX_WRITE32(dev, 0, PX_REG_DMA0);
        PX_WRITE32(dev, 0, PX_REG_DMA1);
        return -EIO;
        }

    /* by setting the end lines, DMA is activated */
    i = (dev->lastline - dev->firstline) / 2; /* lines per field */
    i = dev->xsize*i + dev->dmahole*i; /* total pixels */
    PDEBUG("end dma is %08lx\n",addr+i);
    sprintf(dev->pgmheader, PX_HEADERFMT, dev->xsize, dev->ysize);

    PX_WRITE32(dev, addr + i,                PX_REG_DMAEND0);
    PX_WRITE32(dev, addr + i + dev->dmahole, PX_REG_DMAEND1);

    px_set_mode(dev, PX_REG_GRAB_ON);
    dev->isup = 1; /* reset "just-up" counter */
    return 0;
}

/* This code is common to the open code of various devices */

int px_initialize_operation(Px_Dev *dev, int dmaflag)
{
    int result = 0;
    unsigned long addr;

    px_bringup(dev);

    PDEBUG("initialize, dma is %i\n",dmaflag);
    if (dmaflag && (dev->xsize==0)) { /* provide defaults */
        dev->xsize = 320;
        dev->ysize = 204;
        dev->firstline = 0;
        dev->lastline = 408;
        dev->dmahole = 0;
    }

    /* request the irq */
    if (!dev->irq_active) { /* only on first open */
        result = request_irq(dev->irq, px_interrupt,
                             0 /* slow, unshared */, "px500", dev);
        if (result)
            goto fail_irq;
        dev->irq_active = dev->irq;
        px_set_control_bit(dev, PX_BIT_IRQ_ENABLE);
        px_set_control_bit(dev, PX_BIT_IRQ_ACK); /* paranoid */
    }

    if (dmaflag) {     /* try to allocate the DMA buffer */
        int kb = (dev->xsize*dev->ysize+1023)/1024;

        px_allocate_dma(dev,kb); /* by default */
        if (!dev->dmabuffer) {
            result = -ENOMEM; goto fail_dma;
        }

        addr = virt_to_bus((void *)dev->dmabuffer);
        PDEBUG("dmabuf: virt %p, bus %lx\n",dev->vdmabuffer,addr);

	PDEBUG("xsize4: %i\n",dev->xsize);
        result = px_set_dma_parameters(dev);
        if (result) goto fail_parameters;
    }
    /* else no dmaflag: don't activate dma */
    
    dev->usage++; /* success */
    return 0;

fail_parameters:
    px_free_dma(dev);
fail_dma:
    px_clr_control_bit(dev, PX_BIT_IRQ_ENABLE);
    px_clr_control_bit(dev, PX_BIT_IRQ_ACK);
    free_irq(dev->irq,dev);
    dev->irq_active = 0;
fail_irq:
    if (!dev->usage)
        px_shutdown(dev);
    return result;
}


/*-------------------------------------------------------- open */

int px_open (struct inode *inode, struct file *filp)
{
    int type = TYPE(inode->i_rdev);
    int num = NUM(inode->i_rdev);
    Px_Dev *dev = px_devices;
    int i, result = 0;

    if (type > PX_MAX_TYPE) return -ENODEV;
    for (i=0; dev && i<num; dev = dev->next, i++)
        /* nothing */;
    if (!dev)
        return -ENODEV;
    filp->private_data = dev;
    filp->f_op = px_fop_array[type];

    if (type) /* if type==0, fall through */
      if (filp->f_op->open)
	return (filp->f_op->open)(inode,filp); /* special open code */
      else
	return 0;

    /* do DMA, and all default values */
    dev->xsize = 0; /* force default on */
    result = px_initialize_operation(dev, 1);
    if (result) return result;
    MOD_INC_USE_COUNT;
    return 0;
}

/*--------------------------------------------------other open code */

int px_pgm_open (struct inode *inode, struct file *filp)
{
    Px_Dev *dev = filp->private_data; /* initialized in px_open */
    int result = 0;

    /* do DMA, and default values */
    dev->xsize = 0;
    result = px_initialize_operation(dev, 1);
    if (result) return result;
    MOD_INC_USE_COUNT;
    return 0;
}

int px_ctl_open (struct inode *inode, struct file *filp)
{
    Px_Dev *dev = filp->private_data; /* initialized in px_open */
    int result = 0;

    /* no DMA, and other args are unused */
    result = px_initialize_operation(dev, 0);
    if (result) return result;
    MOD_INC_USE_COUNT;
    return 0;
}

/* hires, common to pgm and non-pgm nodes */
int px_hib_open (struct inode *inode, struct file *filp)
{
    Px_Dev *dev = filp->private_data; /* initialized in px_open */
    int result = 0;

    dev->xsize = 640;
    dev->ysize = 480;
    dev->firstline = 0;
    dev->lastline = 480;
    dev->dmahole = 640;
    result = px_initialize_operation(dev, 1);
    if (result) return result;
    MOD_INC_USE_COUNT;
    return 0;
}


/*-------------------------------------------------------- close */

release_t px_close (struct inode *inode, struct file *filp)
{
    Px_Dev *dev = filp->private_data;

    MOD_DEC_USE_COUNT;
    dev->usage--;
   if (dev->usage==0) {
	px_set_mode(dev, PX_REG_GRAB_NONE);
	if (dev->irq_active) {
	    px_clr_control_bit(dev, PX_BIT_IRQ_ENABLE);	      
	    free_irq(dev->irq_active, dev);
	    dev->irq_active = 0;
	}
        px_free_dma(dev);
        px_shutdown(dev);
    }
    release_return(0);
}

/*-------------------------------------------------------- read */

read_write_t px_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    Px_Dev *dev = filp->private_data; /* the first listitem */

    /*
     * Return a continuous flow of images. f_pos is kept within our limits,
     * because the % operator on a long long needs a lib function :(
     */
    unsigned int size = dev->xsize * dev->ysize;
    unsigned int now = filp->f_pos;

    if (now >= size)
        filp->f_pos = now = (unsigned long)filp->f_pos % size;

    if (now+count > size) count = size-now;

    /* wait for bringup to complete */
    if (dev->isup < 3) {
        PDEBUG("sleep\n");
        interruptible_sleep_on(&dev->queue);
    }

    /* FIXME: should be "copy from io" */
    memcpy_tofs(buf, (void *)(dev->vdmabuffer + now), count); /* Vdma */
    filp->f_pos += count;
    return count;
}

read_write_t px_pgm_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    Px_Dev *dev = filp->private_data; /* the first listitem */

    unsigned int size = dev->xsize * dev->ysize;
    unsigned int now = filp->f_pos;
    int partialcount = 0;

    /* first, the header */

    PDEBUG("pgm read\n");
    if (now < PX_HEADERLEN) {
        partialcount = (count+now >= PX_HEADERLEN) ? PX_HEADERLEN-now : count;
        memcpy_tofs(buf, dev->pgmheader+now, partialcount);
        filp->f_pos += partialcount;
        if (count == partialcount)
            return count; /* nothing more, yet */
        buf += partialcount;
        count -= partialcount;
        now = 17;
    }

    /* then, the image */
                                 
    now -= 17; /* don't count the header */
    if (now >= size)
        return 0;

    if (now+count > size) count = size-now;

    /* wait for bringup to complete */
    if (dev->isup < 3) {
        PDEBUG("sleep\n");
        interruptible_sleep_on(&dev->queue);
    }

    /* FIXME: should be "copy from I/O" */
    memcpy_tofs(buf, (void *)(dev->vdmabuffer + now), count); /* Vdma */
    filp->f_pos += count;
    return count + partialcount;
}

/*-------------------------------------------------------- write */

read_write_t px_write (struct inode *inode, struct file *filp,
                const char *buf, count_t count)
{
#if 0
    Px_Dev *dev = filp->private_data;
#endif
    return -EINVAL;
}

/*-------------------------------------------------------- ioctl */

int px_ioctl (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
    Px_Dev *dev = filp->private_data; /* the first listitem */
    int err=0, size = _IOC_SIZE(cmd); /* the size bitfield in cmd */
    int res = 0;
    #define IOC_BUFSIZE 24 /* at least five integers, for Px_Size */
    static unsigned char localbuf[IOC_BUFSIZE];
    void *karg = localbuf;
    struct Px_Register *kreg = (struct Px_Register *)karg;

    PDEBUG("ioctl: %08x %08lx: nr %i, size %i, dir %i\n",
	    cmd, arg, _IOC_NR(cmd), _IOC_SIZE(cmd), _IOC_DIR(cmd));

    /*
     * extract the type and number bitfields, and don't decode
     * wrong cmds: return EINVAL before verify_area()
     */
    if (_IOC_TYPE(cmd) != PX_IOC_MAGIC) return -EINVAL;
    if (_IOC_NR(cmd) > PX_IOC_MAXNR) return -EINVAL;

    /*
     * the type is a bitmask, and VERIFY_WRITE catches R/W
     * transfers. `Type' is user-oriented, while
     * verify_area is kernel-oriented, so the concept of "read" and
     * "write" is reversed
     */
    if (_IOC_DIR(cmd) & _IOC_READ)
        err = verify_area(VERIFY_WRITE, (void *)arg, size);
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err =  verify_area(VERIFY_READ, (void *)arg, size);
    if (err) return err;

    /* first, retrieve data from userspace */
    if ((_IOC_DIR(cmd) & _IOC_WRITE) && (size <= IOC_BUFSIZE))
        memcpy_fromfs(karg, (void *)arg, size);

    /* and now the big switch... */
    switch(cmd) {
      case PX_IOCHARDRESET:
        mod_use_count_ = 1;
        dev->usage = 1;        
        /* fall through */
      case PX_IOCRESET:
        px_shutdown(dev);
        px_bringup(dev);
        break;

      case PX_IOCCMD:
	if (kreg->reg < 0x10) { /* write */
	    PDEBUGG("ioctl for rd cmd 0x%lx\n",kreg->reg);
            res = px_wr_command(dev, kreg->reg, kreg->val);
        } else { /* read */
            PDEBUGG("ioctl for wr cmd 0x%lx\n",kreg->reg);
            res = px_rd_command(dev, kreg->reg);
        }
        if (res < 0) return res;
        kreg->val = res;
        res = 0;
        break;

      case PX_IOCSREG:
        PX_WRITE32(dev, kreg->val, kreg->reg);
        break;
      case PX_IOCGREG:
        kreg->val = PX_READ32(dev, kreg->reg);
        break;

      case PX_IOCSPORT:
        res = px_set_port_wait(dev, kreg->reg, kreg->val);
        break;
      case PX_IOCGPORT:
        res = px_get_port(dev, kreg->reg);
        if (res < 0) return res;
        kreg->val = res;
        break;

      case PX_IOCSLUT:
        memcpy_fromfs(dev->lut, (void *)arg, size);
        return px_set_lut(dev);
        return 0;

      case PX_IOCGLUT:
        memcpy_tofs((void  *)arg, dev->lut, size);
        return 0;

      case PX_IOCGGAIN:
        res = px_get_gain(dev, (Px_Gain *)karg);
        break;
      case PX_IOCSGAIN:
        res = px_set_gain(dev, (Px_Gain *)karg);
        break;

      case PX_IOCGDEV:
        memcpy_tofs((void  *)arg, dev, size);
        return 0;

      case PX_IOCGSIZE:
        PDEBUG("get size\n");
        ((struct Px_Size *)karg)->xsize = dev->xsize;
        ((struct Px_Size *)karg)->ysize = dev->ysize;
        ((struct Px_Size *)karg)->firstline = dev->firstline;
        ((struct Px_Size *)karg)->lastline = dev->lastline;
        ((struct Px_Size *)karg)->hole = dev->dmahole;
        break;

      case PX_IOCSSIZE:
        dev->xsize = ((struct Px_Size *)karg)->xsize;
        dev->ysize = ((struct Px_Size *)karg)->ysize;
        dev->firstline = ((struct Px_Size *)karg)->firstline;
        dev->lastline = ((struct Px_Size *)karg)->lastline;
        dev->dmahole = ((struct Px_Size *)karg)->hole;
        res = px_initialize_operation(dev, 1);
        if (res) {
            printk(KERN_ERR "px500: failed to change grabbing parameters\n");
            px_shutdown(dev);
            return res;
        }
        break;
        
      case PX_IOCAUTOGAIN:
	if ( (*(int *)karg) > 0 ) {
	    dev->flags |= PX_FLAG_AUTOGAIN;
            px_set_port_wait(dev,PX_PORT_COMPARE,254); /* compare reg */
            dev->autogain_step = *(int *)karg;
            dev->autogain_now = 0;
            PDEBUG("autogain step %i\n",dev->autogain_step);
        }
	else
	    dev->flags &= ~PX_FLAG_AUTOGAIN;
	break;

      default:
	PDEBUG("ioctl: invalid\n");
        return -EINVAL;
    }
    /* finally, copy data to user space and return */
    if (res < 0) return res;
    if ((_IOC_DIR(cmd) & _IOC_READ) && (size <= IOC_BUFSIZE))
        memcpy_tofs((void *)arg, karg, size);
    return 0;

}

/* The binary node, the full featured one */
struct file_operations px_bin_fops = {
    NULL,          /* px_lseek */
    px_read,
    px_write,
    NULL,          /* px_readdir */
    NULL,          /* px_select */
    px_ioctl,
    NULL,          /* px_mmap */
    px_open,
    px_close,
    NULL,          /* px_fsync */
    NULL,          /* px_fasync */
                   /* nothing more, fill with NULLs */
};

/* The pgm node. Similar to binary... */
struct file_operations px_pgm_fops = {
    NULL,          /* px_lseek */
    px_pgm_read,
    px_write,
    NULL,          /* px_readdir */
    NULL,          /* px_select */
    px_ioctl,
    NULL,          /* px_mmap */
    px_pgm_open,
    px_close,
    NULL,          /* px_fsync */
    NULL,          /* px_fasync */
                   /* nothing more, fill with NULLs */
};

/* The control node does no DMA nor read/write */
struct file_operations px_ctl_fops = {
    NULL,        /* no lseek */
    NULL,        /* no read */
    NULL,        /* no write */
    NULL,          /* px_readdir */
    NULL,          /* px_select */
    px_ioctl,
    NULL,          /* px_mmap */
    px_ctl_open,
    px_close,
    NULL,          /* nothing more... */
};

/* The hi-res binary node, with a different open */
struct file_operations px_hib_fops = {
    NULL,          /* px_lseek */
    px_read,
    px_write,
    NULL,          /* px_readdir */
    NULL,          /* px_select */
    px_ioctl,
    NULL,          /* px_mmap */
    px_hib_open,
    px_close,
    NULL,          /* px_fsync */
    NULL,          /* px_fasync */
                   /* nothing more, fill with NULLs */
};

/* The hi-res pgm node. opens like hib and reads as pgm */
struct file_operations px_hip_fops = {
    NULL,          /* px_lseek */
    px_pgm_read,
    px_write,
    NULL,          /* px_readdir */
    NULL,          /* px_select */
    px_ioctl,
    NULL,          /* px_mmap */
    px_hib_open,
    px_close,
    NULL,          /* px_fsync */
    NULL,          /* px_fasync */
                   /* nothing more, fill with NULLs */
};


/*
 * Finally, the module stuff. A free function, then init and cleanup
 */

void px_freethem(void)
{
    Px_Dev *ptr;

    for (ptr = px_devices; ptr; ) {
	px_free_dma(ptr);
	if (ptr->regs)
	  vfree((void *)(ptr->regs));
        ptr = ptr->next;
        kfree(px_devices);
        px_devices = ptr;
    }
}

/*-------------------------------------------------------- modules */
int init_module(void)
{
    int result, i;
    struct Px_Dev *ptr=NULL, **ptrptr;

    /*
     * This will lead to an insmod error if CONFIG_PCI was not set in the
     * target machine.
     */
    if (!pcibios_present()) {
        printk(KERN_ERR "pxdrv: PCI bios needed: can't install\n");
        return -ENODEV;
    }

    /*
     * First of all allocate the device structures
     */

    for (i=0, ptrptr = &px_devices; ; i++) {
        u8 bus, device;

	/* look for the grabber */
        result = pcibios_find_device(0x8086, 0x1223,  i, &bus, &device);
	if (result) {
	    if (result == PCIBIOS_DEVICE_NOT_FOUND) {
	        result = 0; break; /* this is good */
	    } 
	    printk(KERN_ERR "pxdrv: pcibios: %s\n",
		     pcibios_strerror(result));
	    result = -EIO;
	    break;
	}

	/* so, one is there */	
        PDEBUG("found at bus %i device %i\n",bus,device);

        result = -ENOMEM;
        ptr = kmalloc(sizeof(Px_Dev), GFP_KERNEL);
        if (!ptr)
            goto fail_alloc;
        *ptrptr = ptr; /* the next pointer */
        memset(ptr,0,sizeof(Px_Dev));
	ptr->bus = bus;
	ptr->device = device;
        pcibios_read_config_dword(bus, device, 0x10, &ptr->phys_add);
        /* ptr->phys_add &= PCI_BASE_ADDRESS_MEM_MASK; */
        PDEBUG("registers at %x\n",ptr->phys_add);

	if ( (result = px_init(ptr)) ) {
	    if (result == -ENODEV) {
	        PDEBUG("... not a px500\n");
	        kfree(ptr); *ptrptr=NULL;
	        continue; /* not a px500, but try again */
	    }
	    else {
	        PDEBUG("initialization failed\n");
		break;
	    }
	}
	ptrptr = &(ptr->next); /* store here  the next ptr */
    }

    if (!px_devices) {
        printk(KERN_ERR "px500: no device found\n");
        result = -ENODEV;
	goto fail_alloc;
    }
       
    /*
     * Register your major, and accept a dynamic number
     */
    result = register_chrdev(px_major, "px500", px_fop_array[0]);
    if (result < 0) {
        printk(KERN_WARNING "pxdrv: can't get major %d\n",px_major);
	goto fail_alloc;
    }
    if (px_major == 0) px_major = result; /* dynamic */

#ifndef PX_DEBUG
    register_symtab(NULL); /* otherwise, leave global symbols visible */
#endif

    proc_register_dynamic(&proc_root, &px_proc_entry);

    allocator_init();
    printk(KERN_INFO "px500: driver loaded\n");

    return 0; /* succeed */

  fail_alloc:
    px_freethem();
    return result;
}

void cleanup_module(void)
{
    unregister_chrdev(px_major, "px500");
    px_freethem();
    allocator_cleanup();
    proc_unregister(&proc_root, px_proc_entry.low_ino);
    printk(KERN_INFO "px500: driver unloaded\n");
}
